from beet import Context, Function, Predicate
from typing import Any
import json
from mecha import Mecha

mc = Mecha()

def beet_default(ctx: Context):
    generate_advanced_remove_tool(ctx)

    generate_tag_args(ctx)


def append_tag(
    ctx: Context,
    tag: str,
):

    return f"""
execute 
    if items entity @s weapon.mainhand #minecraft:{tag} 
    run data modify storage smithed.crafter:main root.temp.item_tag append value '#minecraft:{tag}'
"""


def generate_tag_args(ctx: Context):
    minecraft_version: str = ctx.meta["minecraft_version"]
    url = f"https://raw.githubusercontent.com/misode/mcmeta/{minecraft_version}-registries/tag/item/data.json"
    item_tags_file = ctx.cache["smithed.crafter"].download(url)
    with open(item_tags_file) as f:
        item_tags: dict[str, list[str]] = json.load(f)

    commands: list[str] = []
    commands.append(
        "data modify storage smithed.crafter:main root.temp.item_tag set value []"
    )
    for item_tag in item_tags:
        command = append_tag(ctx, item_tag)
        commands.append(command)

    commands.append("function #smithed.crafter:event/query_tags")


    ctx.data.functions[
        f"smithed.crafter:v{ctx.project_version}/block/table/crafting/input/query_tags"
    ].append(mc.serialize(mc.parse("\n".join(commands), multiline=True)))


def delete_tool(ctx: Context, item: str, max_damage: int):
    return f"""
execute 
    if data entity @s {{HandItems:[{{id:"minecraft:{item}"}}]}}
    if score $temp1 smithed.data matches {max_damage}.. 
    run function smithed.crafter:v{ctx.project_version}/block/table/crafting/output/clear_input/delete_tool/sub
"""



def generate_advanced_remove_tool(ctx: Context):
    minecraft_version: str = ctx.meta["minecraft_version"]
    url = f"https://raw.githubusercontent.com/misode/mcmeta/{minecraft_version}-summary/item_components/data.json"
    item_components_file = ctx.cache["smithed.crafter"].download(url)
    with open(item_components_file) as f:
        item_components: dict[str, (list[Any] | dict[str,Any])] = json.load(f)

    commands: list[str] = []

    for item, components in item_components.items():
        if isinstance(components, list):
            components_dict = {x["type"]: x["value"] for x in components}
        else:
            components_dict = components

        max_damage = components_dict.get("minecraft:max_damage", None)
        if max_damage:
            command = delete_tool(ctx, item, max_damage)
            commands.append(command)

    ctx.data.functions[
        f"smithed.crafter:v{ctx.project_version}/block/table/crafting/output/clear_input/delete_tool/vanilla"
    ].append(mc.serialize(mc.parse("\n".join(commands), multiline=True)))


from beet import Context

import logging

logger = logging.getLogger(__name__)


def display_broadcasted_pack(ctx: Context):
    """Simply outputs pack details"""

    logger.info("Building %s %s", ctx.project_name, ctx.project_version)


def output_dist(ctx: Context):
    """Dynamically constructs the output directories for Github CI"""

    if not ctx.meta.get("output_release", True):
        logger.info("> Skipped Output")
        return

    _, _, pack_name = ctx.project_id.partition(".")

    if ctx.data:
        ctx.data.save(
            path=(
                ctx.meta["broadcast_directory"]
                / "dist"
                / pack_name
                / f"smithed_{pack_name}_dp.zip"
            ),
            overwrite=True,
        )

    if ctx.assets:
        ctx.assets.save(
            path=(
                ctx.meta["broadcast_directory"]
                / "dist"
                / ctx.project_id.partition(".")[-1]
                / f"smithed_{pack_name}_rp.zip"
            ),
            overwrite=True,
        )



from beet import Context
from beet.contrib.find_replace import find_replace
from beet.contrib.rename_files import rename_files

from .api import generate_api
from .load import generate_load
from .models import ContextualModel, Versioning



def inject_version(ctx: Context):
    opts = ctx.inject(Versioning).opts

    substitution = opts.refactor.dict()
    del substitution["match"]

    # dynamic renames
    ctx.require(
        find_replace(data_pack={"match": opts.refactor.match}, substitute=substitution)
    )
    ctx.require(rename_files(data_pack={"match": opts.refactor.match} | substitution))


def beet_default(ctx: Context):
    """This plugins generates all the versioning requirements that LL needs

    When writing your data pack, you can ignore any versioning you need.
    This plugin will allow you to automatically version any implementation
      alongside with defining APIs via an `@public` at the top of the file.

    It will generate call function for any api route as defined in the config.
    """

    ContextualModel.ctx = ctx  # TODO: use `configurable` in pydantic v2

    # all things for lantern load impl
    ctx.require(generate_load)

    # refactors file names and paths to inject version
    ctx.require(inject_version)

    # we generate api bindings **after** refactoring
    ctx.require(generate_api)

    # we load this afterwards so that dynamic renames don't "touch" it
    ctx.require("beet.contrib.lantern_load.base_data_pack")



from contextlib import suppress

from beet import Advancement, Context, Function, FunctionTag, Generator

from .models import Versioning, VersioningOptions
from .utils import call_if_version_match


def load_tags(
    ctx: Context,
    opts: VersioningOptions,
    dependencies: str,
    enumerate: str,
    resolve: str,
):
    """Sets the Lantern Load tags for deps, enumeration, and resolution"""

    load_tag = ctx.generate(
        opts.lantern_load.tag_path,
        merge=FunctionTag(
            {
                "values": [
                    {"id": f"#{dependencies}", "required": False},
                    f"#{enumerate}",
                    f"#{resolve}",
                ]
            }
        ),
    )

    with ctx.override(generate_namespace="load", generate_prefix=""):
        ctx.generate(
            opts.lantern_load.step, merge=FunctionTag({"values": [f"#{load_tag}"]})
        )


def clear_schedule_loops(scheduled: list[str]):
    """Ensures every schedule loop is disabled"""

    return "".join(f"schedule clear {function}\n" for function in scheduled)


def resolve_func(ctx: Context, opts: VersioningOptions):
    """Resolves the best version"""

    schedules = [ctx.generate.path(path) for path in opts.scheduled_paths.entries()]

    resolve_function = ctx.generate[opts.lantern_load.function_path](
        "resolve",
        Function(
            clear_schedule_loops(schedules)
            + call_if_version_match(
                opts.scoreholder,
                opts.version,
                ctx.generate.path(opts.lantern_load.function_path),
            )
        ),
    )

    return ctx.generate[opts.lantern_load.tag_path](
        "resolve", merge=FunctionTag({"values": [resolve_function]})
    )


def set_version(generate: Generator, opts: VersioningOptions):
    return generate(
        "set_version",
        Function(
            "".join(
                f"scoreboard players set {opts.scoreholder}.{name} load.status {part}\n"
                for name, part in opts.version.named_parts()
            )
            + f"scoreboard players set {opts.scoreholder}.set load.status 1"
        ),
    )


def resolve_advancements(ctx: Context, opts: VersioningOptions):
    """Select packs that match refactor statement"""
    
    query = ctx.query(match=opts.refactor.match, extend=Advancement)
    if not Advancement in query:
        return
    for (path, advancement) in query[Advancement].keys():
        resolve_advancement(advancement, opts)


def resolve_advancement(advancement: Advancement, opts: VersioningOptions):
    """Adds version checking to advancement conditions

    We do this by adding new value checks to the criteria. Within the player
      conditions, we attempt to append our version_check OR edit the existing
      check for the correct scoreholder_part (TODO: make this clearer probably).
    """

    criteria = advancement.data["criteria"]
    for requirement in criteria.values():
        conditions = requirement.setdefault("conditions", {})
        player_conditions = conditions.setdefault("player", [])

        for name, number in opts.version.named_parts():
            scoreholder_part = f"{opts.scoreholder}.{name}"
            version_check = {
                "condition": "minecraft:value_check",
                "value": {
                    "type": "minecraft:score",
                    "target": {
                        "type": "minecraft:fixed",
                        "name": scoreholder_part,
                    },
                    "score": "load.status",
                },
                "range": number,
            }

            for cond in player_conditions:
                with suppress(KeyError, TypeError):
                    if (
                        cond["condition"] == "minecraft:value_check"
                        and cond["value"]["target"]["name"] == scoreholder_part
                    ):
                        cond["range"] = number
                        break

            else:  # only when there's no break
                player_conditions.append(version_check)


def enumerate_func(ctx: Context, opts: VersioningOptions) -> str:
    """The enumeration function which sets the best version

    Enumeration works by chaining calls. First, we need to check the major
      version (or which ever is the highest one). If it's equivalent to our
      already highest version, then we go down to the next step.

    This goes until we've reached the end of each version step.
    """

    # context generators
    generate_function = ctx.generate[opts.lantern_load.function_path]
    generate_tag = ctx.generate[opts.lantern_load.tag_path]
    generate_enumerate = generate_function["enumerate"]

    set_version_path = set_version(generate_enumerate, opts)

    # We iterate through our version parts (we'll use major, minor, patch) **backwards**:
    #  `enumerate` calls major which calls minor which calls patch
    #  Make `enumerate/patch` first, return the path as the `last_step_function`
    #  Then make `enumerate/minor`, using `last_step_function`
    #  Finally, make `enumerate/major`..
    #  At the very end, `enerumate` can use the path of `enumerate/major`!
    last_step_function = None
    for step, value in reversed(opts.version.named_parts()):
        last_step_function = enumerate_step(
            generate_enumerate,
            opts.scoreholder,
            set_version_path,
            step,
            value,
            last_step_function,
        )

    enumerate_function = generate_function(
        "enumerate",
        Function(
            f"scoreboard players reset {opts.scoreholder}.set load.status\n"
            + "".join(
                f"scoreboard players add {opts.scoreholder}.{name} load.status 0\n"
                for name, _ in opts.version.named_parts()
            )
            + f"function {last_step_function}"
        ),
    )

    return generate_tag(
        "enumerate", merge=FunctionTag({"values": [enumerate_function]})
    )


def enumerate_step(
    generate: Generator,
    scoreholder: str,
    set_version_path: str,
    step: str,
    value: int,
    last_step_path: str | None,
):
    """A single step in the enumerate function

    A step cooresponds to a version part (major, minor, etc).
    """

    step_body = (
        "execute "
        f"if score {scoreholder}.{step} load.status matches ..{value} "
        f"unless score {scoreholder}.{step} load.status matches {value} "
        f"run function {set_version_path}"
    )

    if last_step_path is not None:
        step_body += (
            "\n"
            "execute "
            f"unless score {scoreholder}.set load.status matches 1 "
            f"if score {scoreholder}.{step} load.status matches ..{value} "
            f"if score {scoreholder}.{step} load.status matches {value} "
            f"run function {last_step_path}"
        )

    return generate(step, Function(step_body))


def generate_load(ctx: Context):
    """This is the main versioning plugin

    There are 3 parts to integrating lantern load for library support:
     - Basic load tags
        - Sets the step of the lantern load process to run everything else
     - Enumeration
        - Runs through every single pack of different versions
        - Sets the highest version amongst them all
     - Resolution
        - Uses highest version to run the correct load file
        - Disables the rest
    """

    opts = ctx.inject(Versioning).opts

    with ctx.override(generate_namespace=opts.namespace, generate_prefix=""):
        dependencies = ctx.generate[opts.lantern_load.tag_path].path("dependencies")

        # Enumeration starts on the main version part
        #  Then it checks sub-parts if the current part is the same
        enumerate = enumerate_func(ctx, opts)

        # Resolving involves using the highest version and booting off
        #  cooresponding resources while killing off others
        # Advancements get patched to "disable" if version check is **wrong**
        resolve = resolve_func(ctx, opts)
        resolve_advancements(ctx, opts)

        load_tags(ctx, opts, dependencies, enumerate, resolve)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smithed_libraries',
 'smithed_libraries.plugins',
 'smithed_libraries.plugins.versioning']

package_data = \
{'': ['*'],
 'smithed_libraries': ['packs/actionbar/*',
                       'packs/actionbar/data/smithed.actionbar/advancements/impl/vanilla/bed/*',
                       'packs/actionbar/data/smithed.actionbar/advancements/impl/vanilla/container/*',
                       'packs/actionbar/data/smithed.actionbar/functions/impl/*',
                       'packs/actionbar/data/smithed.actionbar/functions/impl/technical/*',
                       'packs/actionbar/data/smithed.actionbar/functions/impl/vanilla/bed/*',
                       'packs/actionbar/data/smithed.actionbar/functions/impl/vanilla/container/*',
                       'packs/actionbar/data/smithed.actionbar/functions/impl/vanilla/container/check_double_chest_lock/*',
                       'packs/actionbar/data/smithed.actionbar/predicates/impl/*',
                       'packs/actionbar/data/smithed.actionbar/tags/blocks/impl/*',
                       'packs/actionbar/data/smithed.actionbar/tags/functions/event/player/*',
                       'packs/crafter-addon/*',
                       'packs/crafter-addon/data/smithed.crafter/tags/functions/event/*',
                       'packs/crafter-addon/data/smithed.crafter_addon/functions/impl/recipes/shaped/*',
                       'packs/crafter-addon/data/smithed.crafter_addon/functions/impl/recipes/shapeless/*',
                       'packs/crafter-addon/data/smithed.crafter_addon/functions/impl/tags/*',
                       'packs/crafter-addon/data/smithed.crafter_addon/predicates/recipes/*',
                       'packs/crafter/*',
                       'packs/crafter/assets/minecraft/models/item/*',
                       'packs/crafter/assets/smithed.crafter/font/*',
                       'packs/crafter/assets/smithed.crafter/lang/*',
                       'packs/crafter/assets/smithed.crafter/models/block/*',
                       'packs/crafter/assets/smithed.crafter/textures/block/*',
                       'packs/crafter/assets/smithed.crafter/textures/gui/*',
                       'packs/crafter/data/smithed.crafter/advancements/technical/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/break/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/input/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/manage_invalids/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/clear_input/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/clear_input/advanced/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/clear_input/delete_tool/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/cursor_check/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/output/shift_click/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/shapeless_recipes/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/crafting/shapeless_recipes/checks/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/block/table/interact/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/entity/player/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/recipes/*',
                       'packs/crafter/data/smithed.crafter/functions/impl/technical/*',
                       'packs/crafter/data/smithed.crafter/functions/zint/technical/*',
                       'packs/crafter/data/smithed.crafter/item_modifiers/impl/*',
                       'packs/crafter/data/smithed.crafter/loot_tables/blocks/*',
                       'packs/crafter/data/smithed.crafter/predicates/block/table/*',
                       'packs/crafter/data/smithed.crafter/predicates/block/table/special_clear/*',
                       'packs/crafter/data/smithed.crafter/predicates/block/table/tags/*',
                       'packs/crafter/data/smithed.crafter/recipes/*',
                       'packs/crafter/data/smithed.crafter/tags/functions/event/*',
                       'packs/crafter/data/smithed.custom_block/tags/functions/event/*',
                       'packs/custom-block/*',
                       'packs/custom-block/data/smithed.custom_block/advancements/impl/technical/*',
                       'packs/custom-block/data/smithed.custom_block/functions/impl/*',
                       'packs/custom-block/data/smithed.custom_block/functions/impl/place/*',
                       'packs/custom-block/data/smithed.custom_block/functions/impl/technical/*',
                       'packs/custom-block/data/smithed.custom_block/tags/blocks/*',
                       'packs/damage/*',
                       'packs/damage/data/smithed.damage/advancements/impl/player/*',
                       'packs/damage/data/smithed.damage/functions/impl/entity/*',
                       'packs/damage/data/smithed.damage/functions/impl/entity/apply/*',
                       'packs/damage/data/smithed.damage/functions/impl/entity/armor/*',
                       'packs/damage/data/smithed.damage/functions/impl/entity/attributes/*',
                       'packs/damage/data/smithed.damage/functions/impl/entity/health/*',
                       'packs/damage/data/smithed.damage/functions/impl/player/*',
                       'packs/damage/data/smithed.damage/functions/impl/technical/*',
                       'packs/damage/data/smithed.damage/tags/entity_types/*',
                       'packs/damage/data/smithed.damage/tags/functions/event/entity/*',
                       'packs/damage/data/smithed.damage/tags/functions/event/player/*',
                       'packs/enchanter/*',
                       'packs/enchanter/assets/smithed.enchanter/font/*',
                       'packs/enchanter/assets/smithed.enchanter/lang/*',
                       'packs/enchanter/assets/smithed.enchanter/textures/gui/*',
                       'packs/enchanter/data/smithed.enchanter/advancements/impl/*',
                       'packs/enchanter/data/smithed.enchanter/functions/impl/*',
                       'packs/enchanter/data/smithed.enchanter/functions/impl/block/enchanter/*',
                       'packs/enchanter/data/smithed.enchanter/functions/impl/technical/*',
                       'packs/item/*',
                       'packs/item/data/smithed.item/advancements/impl/technical/*',
                       'packs/item/data/smithed.item/functions/impl/durability/*',
                       'packs/item/data/smithed.item/functions/impl/durability/calc_mending/*',
                       'packs/item/data/smithed.item/functions/impl/durability/damage/*',
                       'packs/item/data/smithed.item/functions/impl/durability/damage/force/*',
                       'packs/item/data/smithed.item/functions/impl/durability/process/*',
                       'packs/item/data/smithed.item/functions/impl/lore/*',
                       'packs/item/data/smithed.item/functions/impl/technical/*',
                       'packs/item/data/smithed.item/item_modifiers/*',
                       'packs/item/data/smithed.item/predicates/chance/*',
                       'packs/item/data/smithed.item/tags/functions/event/*',
                       'packs/prevent-aggression/*',
                       'packs/prevent-aggression/data/smithed.prevent_aggression/functions/impl/technical/*',
                       'packs/prevent-aggression/data/smithed.prevent_aggression/tags/entity_types/*']}

install_requires = \
['beet>=0.105.0',
 'bolt-expressions>=0.12.2,<0.13.0',
 'bolt>=0.48.0,<0.49.0',
 'mecha>=0.94.0,<0.95.0',
 'pydantic>=2.5.2,<3.0.0']

setup_kwargs = {
    'name': 'smithed-libraries',
    'version': '0.9.1',
    'description': 'All of the Smithed Libraries',
    'long_description': '# Libraries\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Discord](https://img.shields.io/discord/511303648119226382?color=%236d82cc&label=Discord&logo=discord&logoColor=white)](https://discord.gg/gkp6UqEUph)\n\n> Collection of Libraries for the Smithed Ecosystem\n\n## Downloading\n\nYou can download it from [Stable](https://smithed.dev/libraries) | [Nightly](https://nightly.link/Smithed-MC/Libraries/workflows/nightly-build/main/packs.zip)<br/>\nor<br/>\nYou can build it from source using the [beet](https://github.com/mcbeet/beet) via the [Contributing](CONTRIBUTING.md) instructions.\n',
    'author': 'Smithed Council',
    'author_email': 'team@smithed.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import logging
import re

from beet import Context, Function, FunctionTag

from .models import Versioning, VersioningOptions
from .utils import call_if_version_match


logger = logging.getLogger(__name__)

PUBLIC_PAT = re.compile("^#>? @public")


def generate_call(ctx: Context, opts: VersioningOptions, path: str):
    """Generates the actual call function that checks the version"""

    base_path = path.replace(opts.api.implementation_prefix, "")

    version_check = ctx.generate[opts.api.version_check_path](
        base_path, Function(call_if_version_match(opts.scoreholder, opts.version, path))
    )
    api_path = ctx.generate[opts.api.tag_path](
        base_path, FunctionTag({"values": [version_check]})
    )

    logger.info(api_path)


def generate_api(ctx: Context):
    """Generates API calls based on `@public` listed inside of function files

    Note: Must be the very first line of the file

    TODO: Support @public on any line of the function doc comment
            https://github.com/SpyglassMC/Spyglass/wiki/IMP-Doc
    """

    opts = ctx.inject(Versioning).opts

    query = ctx.query(match=opts.api.match, extend=Function)
    if not Function in query:
        return
    for (path, func) in query[Function].keys():
        if func.lines and path is not None and PUBLIC_PAT.match(func.lines[0]):
            generate_call(ctx, opts, path)
